(use srfi-1)
(use srfi-11)
(use srfi-13)
(use srfi-14)
(use srfi-27)
(use gauche.process)
(use gauche.collection)
(use dbm)
(use dbm.gdbm)
(use text.tree)
(use text.tr)
(use rfc.http)
(use rfc.base64)
(use rfc.uri)
(use sxml.ssax)
(use sxml.sxpath)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-actrie lis)
  (update-actrie! (cons #f #f) lis)
  )

(define (update-actrie! act lis)
  (set-car! act lis)
;;  (set-cdr! act (list->regexp lis))
  act)

;; (define (list->regexp lis)
;;   (string->regexp (string-join (map  regexp-quote lis) "|")))

(define (actrie->list act)
  (car act)
  )

;; (define (actrie->regexp act)
;;   (cdr act))

(define (actrie-put! act str)
  (or (member str (car act))
      (update-actrie! act (cons str (actrie->list act))))
  )

(define (actrie-delete! act str)
  (update-actrie! act (delete! str (actrie->list act)))
  )

;; (define (accontains act str)
;;   (let ((match (rxmatch (actrie->regexp act) str)))
;;     (if match
;; 	(lset-adjoin string=?
;; 		     (accontains act (match 'after))
;; 		     (match))
;; 	'()))
;;   )

;; (define (acsplit act str)
;;   (let ((match (rxmatch (actrie->regexp act) str)))
;;     (if match
;; 	(cons (match 'before)
;; 	      (cons (match) (acsplit act (match 'after))))
;; 	(list str))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define *parser-command* "/usr/local/bin/mecab")
(define *parser-encoding* "utf-8")
(define *noun-str* "名詞")

(define (parse string)
  (define (proc inport outport)
    (display string outport)
    (close-output-port outport)
    (parse-parser-output
     (let loop ()
       (let ((line (read-line inport)))
	 (if (eof-object? line)
	     '()
	     (cons line (loop)))))))
  (call-with-process-io *parser-command* proc :encoding *parser-encoding*))

(define (parse-parser-output lines)
  (unfold (lambda (seed)
	    (equal? "EOS" (car seed))) ; exit condition
	  (lambda (seed) (parse-parser-output-line (car seed))) ; seed->elment
	  cdr ; generate new seed
	  lines	; initial seed
	  (lambda (seed) '()))) ; generate tail

(define (parse-parser-output-line line)
  (let* ((halves (string-split line #\tab))
	 (details (string-split (cadr halves) #\,)))
    (cons (car halves) (car details))))

;; (define (extract-nouns str)
;;   (map car
;;        (filter (lambda (pair)
;; 		 (string=? *noun-str* (cdr pair)))
;; 	       (parse str))))

(define (extract-nouns str)
  (filter-map (lambda (pair)
		(and (string=? *noun-str* (cdr pair))
		     (not (defective-noun? (car pair)))
		     (car pair)))
	      (parse str)))

(define (extract-words str)
  (map car (parse str)))

(define zenkaku->hankaku
  (build-transliterator "\uff01-\uff5e"
			"\x21-\x7e"))
(define (string-zenkaku->hankaku str)
  (with-string-io str (cut zenkaku->hankaku)))

(define *prohibited-chars* '(#\w))
(define (defective-noun? word)
  (let ((word (string-zenkaku->hankaku word)))
    (or
     (string-any char-set:punctuation word)
     (any (lambda (ch)
	    (string-every ch word))
	  *prohibited-chars*))
  ))

(define (extract-words+nouns str)
  (let ((lis (parse str)))
    (values (map car lis)
	    (filter-map (lambda (pair)
			  (and (string=? *noun-str* (cdr pair))
			       (car pair)))
			lis))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; (define *db-prefix* "/home/jiro/working/lingo/db-")
(define *db-prefix*
  (string-append (sys-dirname (port-name (current-load-port)))
		 "/db-"))

(define (dbm-creat path)
  (let ((db (dbm-open <gdbm> :path path :rw-mode :create)))
    (dbm-close db)))

(define (dbm-use path)
  (dbm-open <gdbm> :path path :rw-mode :write))

(define (dbm-list db)
  (dbm-for-each db (lambda (key val)
		     (format #t "~a -> ~a\n" key val))))

(define (x->readable x)
  (call-with-output-string
   (lambda (outport)
     (write x outport))))

(define (readable->x string)
  (call-with-input-string string
    (lambda (inport)
      (read inport))))

(define (dbm-get-obj db string)
  (readable->x (dbm-get db string "#f")))

(define (dbm-put-obj! db string obj2)
  (dbm-put! db string (x->readable obj2)))

(define (dbm-push! db key newval)
  (let ((lis (or (dbm-get-obj db key) '())))
    (let ((pair (member newval lis)))
      (if (eq? pair #f)
	  (dbm-put-obj! db key (cons newval lis))))))

(define (dbm-for-each-obj db proc)
  (dbm-for-each db
		(lambda (key val)
		  (proc key (readable->x val)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define *log-db-path* (string-append *db-prefix* "log"))
(define *log-db* #f)

(define (init-log-db)
  (set! *log-db* (dbm-use *log-db-path*))
  (or (dbm-get *log-db* "n-logs" #f)
      (dbm-put! *log-db* "n-logs" "0"))
  (or (dbm-get *log-db* "next-analysis" #f)
      (dbm-put! *log-db* "next-analysis" "0"))
  )
(define (quit-log-db)
  (dbm-close *log-db*))

(define (read-log num)
  (dbm-get *log-db* (number->string num) #f))

(define (add-log str)
  (let ((num (string->number (dbm-get *log-db* "n-logs"))))
    (dbm-put! *log-db* (number->string num) str)
    (dbm-put! *log-db* "n-logs" (number->string (+ 1 num)))
    num))

(define (n-logs)
  (string->number (dbm-get *log-db* "n-logs")))

(define (next-analysis . optional)
  (cond ((null? optional)
	 (string->number (dbm-get *log-db* "next-analysis")))
	(else
	 (dbm-put! *log-db* "next-analysis"
		   (number->string (car optional))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define *term-db-path* (string-append *db-prefix* "terms"))
(define *term-db* #f)
(define *term-actrie* #f)

(define (init-term-db)
  (set! *term-db* (dbm-use *term-db-path*))
  (set! *term-actrie*
	(make-actrie (dbm-map *term-db* (lambda (x _) x)))))
(define (quit-term-db)
  (dbm-close *term-db*))

(define (add-term term log-num)
  (dbm-push! *term-db* term log-num)
  (actrie-put! *term-actrie* term)
  )
(define (add-terms terms log-num)
  (for-each (cut add-term <> log-num) terms))
(define (read-term term)
  (dbm-get-obj *term-db* term))
(define (delete-term term)
  (dbm-delete! *term-db* term)
  (actrie-delete! *term-actrie* term)
  )
(define (terms-for-each proc)
  (dbm-for-each-obj *term-db* proc))

;; (define (extract-known-terms str)
;;   (accontains *term-actrie* str))
(define (extract-known-terms str)
  (lset-intersection string=?
		     (extract-nouns str)
		     (actrie->list *term-actrie*)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define *rare-rate* 1/800)
(define *frequent-rate* 1/50)
(define *gc-interval* 60)

(define (gc-terms)
  (cond ((>= (- (n-logs) (next-analysis)) *gc-interval*)
	 (gc-terms-real)
	 (next-analysis (n-logs)))))
	 
(define (gc-terms-real)
  (terms-for-each (lambda (key val)
		    (let ((rate #?=(exact->inexact (/ (length val) (n-logs)))))
		      (if (or (< rate *rare-rate*)
			      (< *frequent-rate* rate)
			      (defective-noun? key))
			  (begin
			    (format #t ";; Deleted term: ~a\n" key)
			    (delete-term key))
			  (format #t ";; Retained term: ~a\n" key)
			  )))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; (define (unreferenced-logs)
;;   (let* ((n (n-logs))
;; 	 (v (make-vector n #f))
;; 	 (lis '()))
;;     (terms-for-each (lambda (key val)
;; 		      (for-each (lambda (lognum)
;; 				  (vector-set! v lognum #t))
;; 				val)))
;;     (dotimes (i n)
;;       (if (not (vector-ref v i))
;; 	  (push! lis i)))
;;     lis))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-response terms)
  (generate-by-markov (collect-logs terms)))

(define (collect-logs terms)
  (fold (lambda (term tset)
	  (fold (lambda (logn nset)
		  (lset-adjoin eq? nset logn))
		tset
		(read-term term)
		))
	'()
	terms
	))

(define *n-trial* 10)

(define (rep n proc)
  (unfold (lambda (x) (= x n)) ; last condition
	  (lambda (_) (proc)) ; mapper
	  (lambda (x) (+ x 1)) ; next
	  0 ; initial seed
	  ))

(define (generate-by-markov logs)
  (let ((ht (make-hash-table 'equal?))
	(log-count (length logs)))
    (let ((term-lis (fold (lambda (lmn stat)
			    (let-values ([(words nouns)
					  (extract-words+nouns (read-log lmn))])
			      (add-to-ht-list ht words)
			      (append nouns stat)))
			  '()
			  logs)
		    )
	  (norm-lis (fold (lambda (lmn stat)
			    (append (extract-nouns (read-log lmn)) stat))
			  '()
			  (random-logs log-count))))
      (tree->string
       (select-max-evaluation (cut evaluate-response <> term-lis norm-lis)
			      (let ((lis 
				     (rep *n-trial*
					  (cut generate-by-markov-real ht))))
				;; (format #t ";; ~s\n" lis)
				lis))))))

(define (bayesian-prob response-nouns term-lis norm-lis)
  (let ((term-count (length term-lis))
	(norm-count (length norm-lis)))
    (fold (lambda (lmn stat)
	    (* (/ stat ; P(A)
		  (max 1/100
		       (/ (count (cut string=? lmn <>) norm-lis)
			  norm-count))) ; P(B)
	       (max 1/100
		    (/ (count (cut string=? lmn <>) term-lis)
		       term-count)))) ; P(B|A)
	  1/2
	  response-nouns)))

(define (count-duplicate-excess lis)
  (let ((ht (make-hash-table 'string=?)))
    (fold (lambda (lmn count)
	    (if (hash-table-exists? ht lmn)
		(+ 1 count)
		(begin
		  (hash-table-put! ht lmn #t)
		  count)))
	  0
	  lis
	  )))

(define (evaluate-response response term-lis norm-lis)
  (let ((response-nouns (filter (lambda (word)
				  (or (member word term-lis)
				      (member word norm-lis)))
				response)))
    (/
     ;; Step 1: Bayesian statistics
     ;; P(A|B) = [P(A)/P(B)]P(B|A)
     (bayesian-prob response-nouns term-lis norm-lis)
     
     ;; Step 2: duplicate nouns are bad
     (expt 10 (count-duplicate-excess response))
     )))

(define (random-logs n)
  (let ((supn (n-logs)))
    (rep n (cut random-integer supn))))

(define (select-max-evaluation proc lis)
  (receive (maxev maxel)
	   (fold2 (lambda (el maxev maxel) ; (fold2 proc knil1 knil2 lis)
		    (let ((v (proc el)))
		      (if (or (not maxev)
			      (< maxev v))
			  (values v el)
			  (values maxev maxel))))
		  #f #f
		  lis)
	   #?=(exact->inexact maxev)
	   maxel))

(define (add-to-ht-list ht lis)
  (pair-for-each (lambda (pair)
		   (hash-table-push! ht
				     (car pair)
				     (if (null? (cdr pair))
					 #f
					 (cadr pair))))
		 (cons #t lis)))

(define (generate-by-markov-real ht)
  (cdr
   (let sub ((cur #t))
     (if cur
	 (let ((post (random-select (hash-table-get ht cur))))
	   (cons cur (sub post)))
	 '()))))

(define (random-select lis)
  (list-ref lis (random-integer (length lis))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (lingo-init)
  (random-source-randomize! default-random-source)
  (init-log-db)
  (init-term-db))

(define (lingo-quit)
  (quit-log-db)
  (quit-term-db))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (study-message-real str)
  (let ((num (add-log str))
	(terms (extract-nouns str)))
    (add-terms terms num)))

(define (study-message str)
  (cond ((string-null? str))
	(else
	 (gc-terms)
	 (study-message-real str))))

(define (study-file filename)
  (with-input-from-file filename
    (cut port-for-each study-message read-line)))

;; (define (respond-random)
;;   (make-response (list (random-select (actrie->list *term-actrie*))))
;;   )
(define (respond-random)
  (make-response #?=(rep 3 
			 (cute random-select (actrie->list *term-actrie*)))))

(define (respond-to-real str)
  (let ((terms #?=(extract-known-terms str)))
    (if (null? terms)
	(respond-random)
	(make-response terms))))

(define (respond-to str)
  (if (string-null? str)
      ""
      (let ((resp (respond-to-real str)))
	(study-message str)
	resp)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (interactive-loop-real)
  (display "lingo> ")
  (flush)
  (let ((line (read-line)))
    (cond ((eof-object? line)
	   (newline))
	  ((string-null? line)
	   (display "type Ctrl-D to end conversation.\n")
	   (interactive-loop-real))
	  ((not (eof-object? line))
	   (format #t "~a\n" (respond-to line))
	   (interactive-loop-real)))))

(define (interactive-loop)
  (lingo-init)
  (interactive-loop-real)
  (lingo-quit))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (authorization-value user passwd)
  (string-append "Basic "
		 (base64-encode-string (string-append user ":" passwd))))

(define *screen-name* "screen_name")
(define *password* "password")
(define *server* "api.twitter.com")
(define *max-chars* 140)

(define *taboo-prefix-list* '("@" "＠" "#" "＃" "http://"))
(define *taboo-word-list*
  (map symbol->string '(RT QT ON OFF STOP QUIT End Cancel Arret Unsubscribe FOLLOW LEAVE FAV RT SET LOCATION WHOIS GET FAV STATS INVITE)))

(define (tw-strip-commands str)
  (let* ((str0
	  (regexp-replace-all #/[[:graph:]]+(?=[^[:graph:]]|$)/ str ;|)
			      (lambda (match)
				(if (or (any (cut string-prefix? <> (match))
					     *taboo-prefix-list*)
					(any (cut string=? <> (match))
					     *taboo-word-list*))
				    ""
				    (match)
				    ))))
	 (str1 ;; bad ad-hoc technique
	  (regexp-replace-all #/@[a-zA-Z0-9_]*/ str0 "")))
    str1))

(define (tw-strip-length str)
  (if (< *max-chars* (string-length str))
      (string-append (substring str 0 (- *max-chars* 1)) "$")
      str))

(define (tw-stub thunk)
  (let-values ([(code headers body) (thunk)])
    (if (string=? "200" code)
	body
	(cons code body))))

(define (tw-update status)
  (tw-stub
   (cut http-post *server*
	"/1/statuses/update.xml"
	(string-append "status="
		       (uri-encode-string status))
	:content-type "application/x-www-form-urlencoded"
	:authorization (authorization-value *screen-name*
					    *password*))))

(define (tw-home-timeline . maybe-since-id)
  (tw-stub
   (cut http-get *server*
	(tree->string `("/1/statuses/home_timeline.xml"
			,(if (null? maybe-since-id)
			     '()
			     `("?since_id="
			       ,(if (string? (car maybe-since-id))
				    (car maybe-since-id)
				    (x->string (car maybe-since-id)))))))
	:authorization (authorization-value *screen-name*
					    *password*))))

(define (tw-friends-ids)
  (tw-stub
   (cut http-get *server*
	(string-append "/1/friends/ids.xml?screen_name=" *screen-name*)
	:authorization (authorization-value *screen-name*
					    *password*))))

(define (tw-followers-ids)
  (tw-stub
   (cut http-get *server*
	(string-append "/1/followers/ids.xml?screen_name=" *screen-name*)
	:authorization (authorization-value *screen-name*
					    *password*))))

(define (tw-friendships-create id)
  (tw-stub
   (cut http-post *server*
	(string-append "/1/friendships/create.xml?id=" (x->string id))
	""
	:authorization (authorization-value *screen-name*
					    *password*))))

(define (tw-friendships-destroy id)
  (tw-stub
   (cut http-post *server*
	(string-append "/1/friendships/destroy.xml?id=" (x->string id))
	""
	:authorization (authorization-value *screen-name*
					    *password*))))

(define (string->sxml str)
  (call-with-input-string str (cut ssax:xml->sxml <> '())))

(define (tw-display-error cell)
  (format #t ";; An error occurred:\n;; HTTP response code: ~a\n" (car cell))
  ;; (let ((sxml (string->sxml (cdr cell))))
  ;;   (format #t ";; Request: ~a\n" (car ((sxpath '(// request *text*)) sxml)))
  ;;   (format #t ";; Error: ~a\n" (car ((sxpath '(// error *text*)) sxml))))
  (format #t ";; The exact body twitter sent is as follows:\n~a\n" (cdr cell))
  )

(define *last-id-path* (string-append *db-prefix* "last-id"))

(define (process-home-timeline)
  (let ((maybe-xml
	 (apply tw-home-timeline
		(if (file-exists? *last-id-path*)
		    (list (with-input-from-file *last-id-path* (cut read)))
		    '()))))
    (if (pair? maybe-xml) ; error
	(tw-display-error maybe-xml)
	(let* ((sxml (string->sxml maybe-xml))
	       (status-list ((sxpath '(statuses status)) sxml))
	       )
	  (cond ((null? status-list)) ; do nothing
		(else
		 (for-each process-a-status status-list)
		 (with-output-to-file *last-id-path* 
		   (cut write (last-status-id status-list)))))))))

(define (take-element path tree)
  (car ((sxpath path) tree)))

(define (last-status-id status-list)
  ;; (car ((sxpath '(id *text*)) (car status-list)))
  (take-element '(id *text*) (car status-list))
  )

(define (process-a-status status-tree)
  (with-error-handler
   (cut report-error <>)
   (lambda ()
     (let ((text (take-element '(text *text*) status-tree)))
       (cond ((string=? *screen-name*
			(take-element '(user screen_name *text*) status-tree))
	      ;; (format #t "(Ign) ~a\n" text)
	      )
	     ((not (string-any #[ぁ-ん] text))
	      )
	     ((let ((lis
		     ((sxpath '(in_reply_to_screen_name *text*)) status-tree)))
		(and (not (null? lis))
		     (string=? *screen-name* (car lis))))
	      (format #t "(<<<) ~a\n" text)
	      (post-response (tw-strip-commands text)
			     (take-element '(user screen_name *text*) status-tree)
			     )
	      )
	     (else
	      ;; (format #t "(Std) ~a\n" text)
	      (study-message (tw-strip-commands text)))
	     )))))

(define (post-real str)
  (let ((maybe-body (tw-update str)))
    (if (pair? maybe-body)
	(tw-display-error maybe-body)
	(format #t "(>>>) ~a\n" str))))

(define (post-random)
  (post-real (tw-strip-length (tw-strip-commands (respond-random)))))

(define (post-response text screen-name)
  (let* ((str
	  (tw-strip-length (string-append "@" screen-name " "
					  (tw-strip-commands (respond-to text))))))
    (post-real str)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (update-friendship)
  (let ((maybe-friends-xml (tw-friends-ids))
	(maybe-followers-xml (tw-followers-ids)))
    (cond ((pair? maybe-friends-xml)
	   (tw-display-error maybe-friends-xml))
	  ((pair? maybe-followers-xml)
	   (tw-display-error maybe-followers-xml))
	  (else
	   (let ((friends-ids ((sxpath '(ids id *text*))
			       (string->sxml maybe-friends-xml)))
		 (followers-ids ((sxpath '(ids id *text*))
				 (string->sxml maybe-followers-xml))))
	     (for-each (lambda (new-friend)
			 (let ((p (tw-friendships-create new-friend)))
			   (if (pair? p)
			       (tw-display-error p)
			       (format #t ";; Started following user #~a.\n"
				       new-friend)
			       )))
		       (lset-difference equal? followers-ids friends-ids))
	     (for-each (lambda (lost-friend)
			 (let ((p (tw-friendships-destroy lost-friend)))
			   (if (pair? p)
			       (tw-display-error p)
			       (format #t ";; Stopped following user #~a.\n"
				       lost-friend)
			       )))
		       (lset-difference equal? friends-ids followers-ids)))))))
